@@ -1,4 +1,5 @@
 #!/usr/bin/python3
+# coding: UTF-8
 '''Mock D-BUS objects for test suites.'''
 
 # This program is free software; you can redistribute it and/or modify it under
@@ -19,11 +20,10 @@ import unittest
 import subprocess
 import signal
 import os
+import errno
 
 import dbus
-import dbus.mainloop.glib
 import dbus.service
-from gi.repository import GObject
 
 # global path -> DBusMockObject mapping
 objects = {}
@@ -49,7 +49,7 @@ class DBusMockObject(dbus.service.Object):
         logfile: When given, method calls will be logged into that file name;
                  if None, logging will be written to stdout.
         '''
-        super().__init__(bus_name, path)
+        super(self.__class__, self).__init__(bus_name, path)
 
         self.bus_name = bus_name
         self.interface = interface
@@ -67,7 +67,6 @@ class DBusMockObject(dbus.service.Object):
 
     def __del__(self):
         if self.logfile:
-            print('DBusMockObject __del__ closing log')
             self.logfile.close()
 
     @dbus.service.method(dbus.PROPERTIES_IFACE,
@@ -122,7 +121,7 @@ class DBusMockObject(dbus.service.Object):
         path: D-Bus object path
         interface: Primary D-Bus interface name of this object (where
                    properties and methods will be put on)
-        properties: A property_name (string) → property (Variant) map with initial
+        properties: A property_name (string) → value map with initial
                     properties on "interface"
         methods: An array of 4-tuples (name, in_sig, out_sig, code) describing
                  methods to add to "interface"; see AddMethod() for details of
@@ -183,9 +182,13 @@ class DBusMockObject(dbus.service.Object):
                  that returns a string; use '' for methods that do not return
                  anything.
         code: Python 3 code to run in the method call; you have access to the
-              arguments through the args list, and can set the return value by
-              assigning a value to the "ret" variable. When specifying '', the
-              method will not do anything (except logging) and return None.
+              arguments through the "args" list, and can set the return value
+              by assigning a value to the "ret" variable. You can also read the
+              global "objects" variable, which is a dictionary mapping object
+              paths to DBusMockObject instances.
+              
+              When specifying '', the method will not do anything (except
+              logging) and return None.
         '''
         if not interface:
             interface = self.interface
@@ -202,7 +205,7 @@ class DBusMockObject(dbus.service.Object):
         # check in dbus-python; we need to set it manually instead
         dbus_method = dbus.service.method(interface,
                                           out_signature=out_sig)(method)
-        dbus_method.__name__ = name
+        dbus_method.__name__ = str(name)
         dbus_method._dbus_in_signature = in_sig
         dbus_method._dbus_args = ['arg%i' % i for i in range(1, n_args + 1)]
 
@@ -250,6 +253,47 @@ class DBusMockObject(dbus.service.Object):
             pass
         self.props.setdefault(interface, {})[name] = value
 
+    @dbus.service.method('org.freedesktop.DBus.Mock',
+                         in_signature='sa{sv}',
+                         out_signature='')
+    def AddProperties(self, interface, properties):
+        '''Add several properties to this object
+
+        interface: D-Bus interface to add this to. For convenience you can
+                   specify '' here to add the property to the object's main
+                   interface (as specified on construction).
+        properties: A property_name (string) → value map
+        '''
+        for k, v in properties.items():
+            self.AddProperty(interface, k, v)
+
+    @dbus.service.method('org.freedesktop.DBus.Mock',
+                         in_signature='sssav',
+                         out_signature='')
+    def EmitSignal(self, interface, name, signature, args):
+        '''Emit a signal from the object.
+
+        interface: D-Bus interface to send the signal from. For convenience you
+                   can specify '' here to add the method to the object's main
+                   interface (as specified on construction).
+        name: Name of the signal
+        signature: Signature of input arguments; for example "ias" for a signal
+                that takes an int32 and a string array as arguments; see
+                http://dbus.freedesktop.org/doc/dbus-specification.html#message-protocol-signatures
+        args: variant array with signal arguments; must match order and type in
+              "signature"
+        '''
+        if not interface:
+            interface = self.interface
+
+        fn = lambda self, *args: self.log('emit %s.%s: %s' % (interface, name, args))
+        fn.__name__ = str(name)
+        dbus_fn = dbus.service.signal(interface)(fn)
+        dbus_fn._dbus_signature = signature
+        n_args = len(dbus.Signature(signature))
+        dbus_fn._dbus_args = ['arg%i' % i for i in range(1, n_args + 1)]
+        dbus_fn(self, *args)
+
     def mock_method(self, interface, dbus_method, *args, **kwargs):
         '''Master mock method.
 
@@ -300,7 +344,7 @@ class DBusMockObject(dbus.service.Object):
                 mock_interfaces.setdefault(interface, {})[method] = self.methods[interface][method][3]
         self._dbus_class_table[cls] = mock_interfaces
 
-        xml = super().Introspect(object_path, connection)
+        xml = super(self.__class__, self).Introspect(object_path, connection)
 
         # restore original class table
         self._dbus_class_table[cls] = orig_interfaces
@@ -326,7 +370,7 @@ class DBusTestCase(unittest.TestCase):
         
         This gets stopped automatically in tearDownClass().
         '''
-        (klass.session_bus_pid, addr) = klass.start_dbus()
+        (DBusTestCase.session_bus_pid, addr) = klass.start_dbus()
         os.environ['DBUS_SESSION_BUS_ADDRESS'] = addr
 
     @classmethod
@@ -335,21 +379,21 @@ class DBusTestCase(unittest.TestCase):
         
         This gets stopped automatically in tearDownClass().
         '''
-        (klass.system_bus_pid, addr) = klass.start_dbus()
+        (DBusTestCase.system_bus_pid, addr) = klass.start_dbus()
         os.environ['DBUS_SYSTEM_BUS_ADDRESS'] = addr
 
     @classmethod
     def tearDownClass(klass):
         '''Stop private session/system buses'''
 
-        if klass.session_bus_pid is not None:
-            klass.stop_dbus(klass.session_bus_pid)
+        if DBusTestCase.session_bus_pid is not None:
+            klass.stop_dbus(DBusTestCase.session_bus_pid)
             del os.environ['DBUS_SESSION_BUS_ADDRESS']
-            klass.session_bus_pid = None
-        if klass.system_bus_pid is not None:
-            klass.stop_dbus(klass.system_bus_pid)
+            DBusTestCase.session_bus_pid = None
+        if DBusTestCase.system_bus_pid is not None:
+            klass.stop_dbus(DBusTestCase.system_bus_pid)
             del os.environ['DBUS_SYSTEM_BUS_ADDRESS']
-            klass.system_bus_pid = None
+            DBusTestCase.system_bus_pid = None
 
     @classmethod
     def start_dbus(klass):
@@ -376,11 +420,22 @@ class DBusTestCase(unittest.TestCase):
         start_system_bus() and start_session_bus(), these buses are
         automatically stopped in tearDownClass().
         '''
-        os.kill(pid, signal.SIGTERM)
-        try:
-            os.waitpid(pid, 0)
-        except OSError:
-            pass
+        signal.signal(signal.SIGTERM, signal.SIG_IGN)
+        timeout = 50
+        while timeout > 0:
+            try:
+                os.kill(pid, signal.SIGTERM)
+            except OSError as e:
+                if e.errno == errno.ESRCH:
+                    break
+                else:
+                    raise
+            time.sleep(0.1)
+        else:
+            sys.stderr.write('ERROR: timed out waiting for bus process to terminate\n')
+            os.kill(pid, signal.SIGKILL)
+            time.sleep(0.5)
+        signal.signal(signal.SIGTERM, signal.SIG_DFL)
 
     @classmethod
     def get_dbus(klass, system_bus=False):
@@ -454,6 +509,14 @@ class DBusTestCase(unittest.TestCase):
 
         return daemon
 
+# Python 2 backwards compatibility
+if sys.version_info[0] < 3:
+    import re
+    def assertRegex(self, value, pattern):
+        if not re.search(pattern, value):
+            raise self.failureException('%r not found in %s' % (pattern, value))
+    DBusTestCase.assertRegex = assertRegex
+
 
 def parse_args():
     parser = argparse.ArgumentParser(description='mock D-BUS object')
@@ -471,6 +534,9 @@ def parse_args():
 
 
 if __name__ == '__main__':
+    import dbus.mainloop.glib
+    from gi.repository import GObject
+
     args = parse_args()
     dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
     bus_name = dbus.service.BusName(args.name,
@@ -480,4 +546,5 @@ if __name__ == '__main__':
                                     do_not_queue=True)
 
     main_object = DBusMockObject(bus_name, args.path, args.interface, {}, args.logfile)
+    objects[args.path] = main_object
     GObject.MainLoop().run()
